/* See the file Distribution for distribution terms.
	Copyright (c) 1994-1995 Ari Halberstadt */

/*	This file contains an interface to Thread Library that simulates Apple's
	Thread Manager. This interface allows you to easily switch between Apple's
	Thread Manager and my Thread Library. All routines and actions supported
	by Apple's Thread Manager, version 2.0, are supported by the routines in
	this library, with the exception of preemptive threads and thread pools.
	Operations on preemptive threads result in no-ops, and attempting to create
	a preemptive thread results in a threadProtocolErr, just as it would
	for the PowerPC version of Thread Manager. Since preemptive threads
	are not supported, thread pools are less important. Threads are always
	created using NewPtr and disposed of using DisposePtr; requests to
	create a thread from a pool, or to recycle a thread when it is
	disposed of, are ignored.
	
	The routines in this library have the same syntax as Thread Manager
	routines, and the names of the routines are the same except for the
	prefix "TLM". For instance, to create a new thread, you would use
	TLMNewThread instead of NewThread. The header file "ThreadLibraryManager.h"
	will automatically redefine Thread Manager routines to use the "TLM"
	routines in this file.
	
	If Thread Manager is available, the routines in this file call directly
	to Thread Manager. If Thread Manager is not available, then my Thread
	Library is used. This provides your application with automatic
	compatability with previous and future versions of MacOS. On systems
	on which Thread Manager is not installed (i.e., systems prior to System 7.5
	on which the Thread Manager extension has not been installed) Thread
	Library is used to simulate its operation. When Copland (System 8)
	redefines process architecture, your application will automatically use
	the new version of Thread Manager, and not my Thread Library (which may be
	incompatible with Copland).
	
	Warning: I have not tested every routine in this file. Use at your own risk. */

/* Prevent redefinition of routines in "ThreadLibraryManager.h". */
#define THREAD_MANAGER_AS_THREAD_LIBRARY (0)
#define THREAD_LIBRARY_AS_THREAD_MANAGER (0)

#include <ConditionalMacros.h>
#include <Gestalt.h>
#include <MixedMode.h>
#include <OSUtils.h>
#include "ThreadLibrary.h"
#include "ThreadLibraryManager.h"

/*	Macros for making callbacks. CFM68K uses routine descriptors, while
	PPC and old-style 68K use function pointers. */
#if GENERATING68K && GENERATINGCFM
	#define CallEntryProc(proc, param)							((void *) CallUniversalProc(proc, uppThreadEntryProcInfo, param))
	#define CallSchedulerProc(proc, info)						((ThreadID) CallUniversalProc(proc, uppThreadSchedulerProcInfo, info))
	#define CallSwitcherProc(proc, thread, param)			((void) CallUniversalProc(proc, uppThreadSwitchProcInfo, thread, param))
	#define CallTerminatorProc(proc, thread, param)			((void) CallUniversalProc(proc, uppThreadTerminationProcInfo, thread, param))
	#define CallDebuggerNewThreadProc(proc, thread)			((void) CallUniversalProc(proc, uppDebuggerNewThreadProcInfo, thread))
	#define CallDebuggerDisposeThreadProc(proc, thread)	((void) CallUniversalProc(proc, uppDebuggerDisposeThreadProcInfo, thread))
	#define CallDebuggerSchedulerProc(proc, info)			CallUniversalProc(proc, uppDebuggerThreadSchedulerProcInfo, info)
#else
	#define CallEntryProc(proc, param)							proc(param)
	#define CallSwitcherProc(proc, thread, param)			proc(thread, param)
	#define CallTerminatorProc(proc, thread, param)			proc(thread, param)
	#define CallSchedulerProc(proc, info)						proc(info)
	#define CallDebuggerNewThreadProc(proc, thread)			proc(thread)
	#define CallDebuggerDisposeThreadProc(proc, thread)	proc(thread)
	#define CallDebuggerSchedulerProc(proc, info)			proc(info)
#endif

/*	Global state. */
static struct {
	ThreadSchedulerProcPtr threadScheduler;
	DebuggerNewThreadProcPtr notifyNewThread;
	DebuggerDisposeThreadProcPtr notifyDisposeThread;
	DebuggerThreadSchedulerProcPtr notifyThreadScheduler;
} gThread;

/*	Additional information maintained for each thread. */
typedef struct {
	ThreadType sn;
	ThreadID id;
	
	ThreadEntryProcPtr entry;
	void *entryParam;
	void **entryResult;
	
	ThreadSwitchProcPtr switcherIn;
	void *switcherInParam;
	
	ThreadSwitchProcPtr switcherOut;
	void *switcherOutParam;
	
	ThreadTerminationProcPtr terminator;
	void *terminatorParam;
	
	long critical;
} ThreadManagerType;

/*	Create the main thread if it doesn't already exist. */
static OSErr CreateMainThread(void)
{
	ThreadManagerType *thread;
	ThreadType main;
	OSErr err;

	/* check if main thread exists */
	main = ThreadMain();
	err = ThreadError();
	if (! err && main == THREAD_NONE) {

		/* create main thread */
		thread = (ThreadManagerType *) NewPtrClear(sizeof(ThreadManagerType));
		if (! thread)
			err = (MemError() ? MemError() : memFullErr);
		if (! err) {
			thread->sn = ThreadBeginMain(NULL, NULL, thread);
			err = ThreadError();
			if (! err)
				thread->id = kApplicationThreadID;
		}
		if (err)
			DisposePtr((Ptr) thread);
	}
	return(err);
}

/*	Convert a Thread Manager ID to a Thread Library serial number. */
static OSErr ThreadIDToSN(ThreadID id, ThreadType *sn)
{
	OSErr err;
	
	err = noErr;
	*sn = THREAD_NONE;
	switch (id) {
	case kNoThreadID:
		*sn = THREAD_NONE;
		break;
	case kCurrentThreadID:
		*sn = ThreadActive();
		break;
	case kApplicationThreadID:
		err = CreateMainThread();
		if (! err)
			*sn = ThreadMain();
		break;
	default:
		*sn = id - 1;
		break;
	}
	return(err);
}

/*	Convert a Thread Library serial number to a Thread Manager ID. */
static OSErr ThreadSNToID(ThreadType sn, ThreadID *id)
{
	OSErr err;

	err = CreateMainThread();
	if (! err) {
		if (sn == THREAD_NONE)
			*id = kNoThreadID;
		else if (sn == ThreadMain())
			*id = kApplicationThreadID;
		else
			*id = sn + 1;
	}
	/* assert: *id != kCurrentThreadID */
	return(err);
}

/*	Convert a Thread Library serial number to a pointer to our thread data. */
static OSErr ThreadSNToPtr(ThreadType sn, void **p)
{
	*p = ThreadData(sn);
	return(ThreadError());
}

/*	Convert a Thread Manager ID to a pointer to our thread data. */
static OSErr ThreadIDToPtr(ThreadID id, void **p)
{
	ThreadType sn;
	OSErr err;
	
	err = ThreadIDToSN(id, &sn);
	if (! err)
		err = ThreadSNToPtr(sn, p);
	return(err);
}

/*	Function called by Thread Library which in turn calls a custom scheduler. */
static ThreadType tlm_scheduler(ThreadType suggested)
{
	SchedulerInfoRec info;
	ThreadType active;
	OSErr err;
	
	if (gThread.threadScheduler) {
		info.InfoRecSize = sizeof(SchedulerInfoRec);
		info.InterruptedCoopThreadID = kNoThreadID;
		active = ThreadActive();
		err = ThreadError();
		if (! err) {
			err = ThreadSNToID(active, &info.CurrentThreadID);
			if (! err) {
				err = ThreadSNToID(suggested, &info.SuggestedThreadID);
				if (! err) {
					info.SuggestedThreadID = CallSchedulerProc(gThread.threadScheduler, &info);
					if (gThread.notifyThreadScheduler)
						info.SuggestedThreadID = CallDebuggerSchedulerProc(gThread.notifyThreadScheduler, &info);
					err = ThreadIDToSN(info.SuggestedThreadID, &suggested);
				}
			}
		}
		if (err)
			suggested = THREAD_NONE;
	}
	return(suggested);
}

/*	Callback function called by Thread Library which in turn calls a function
	using Thread Manager calling conventions. */
static void tlm_begin(ThreadType sn, void *data)
{
	#pragma unused (sn)
	ThreadManagerType *thread = data;
	
	if (gThread.notifyNewThread)
		CallDebuggerNewThreadProc(gThread.notifyNewThread, thread->id);
}

/*	Callback function called by Thread Library which in turn calls a function
	using Thread Manager calling conventions. */
static void tlm_end(ThreadType sn, void *data)
{
	#pragma unused (sn)
	ThreadManagerType *thread = data;
	
	if (thread->terminator)
		CallTerminatorProc(thread->terminator, thread->id, thread->terminatorParam);
	if (gThread.notifyDisposeThread)
		CallDebuggerDisposeThreadProc(gThread.notifyDisposeThread, thread->id);
}

/*	Callback function called by Thread Library which in turn calls a function
	using Thread Manager calling conventions. */
static void tlm_suspend(void *data)
{
	ThreadManagerType *thread = data;
	
	if (thread->switcherOut)
		CallSwitcherProc(thread->switcherOut, thread->id, thread->switcherOutParam);
}

/*	Callback function called by Thread Library which in turn calls a function
	using Thread Manager calling conventions. */
static void tlm_resume(void *data)
{
	ThreadManagerType *thread = data;
	
	if (thread->switcherIn)
		CallSwitcherProc(thread->switcherIn, thread->id, thread->switcherInParam);
}

/*	Callback function called by Thread Library which in turn calls a function
	using Thread Manager calling conventions. */
static void tlm_entry(void *data)
{
	ThreadManagerType *thread = data;
	void *result;
	
	if (thread->entry) {
		result = CallEntryProc(thread->entry, thread->entryParam);
		if (thread->entryResult)
			*thread->entryResult = result;
	}
}

/*	Ascertains if Thread Manager is available. If this returns true, all
	routines call Thread Manager. If this returns false, all routines
	use Thread Library. */
static Boolean ThreadManagerIsAvailable(void)
{
	static short available = -1;
	long response;
	
	if (available == -1) {
		if (Gestalt(gestaltThreadMgrAttr, &response) != noErr)
			available = false;
		else {
			available = ((response & (1UL << gestaltThreadMgrPresent)) != 0);
			#if GENERATINGCFM
				if (available)
					available = (NewThread != NULL);
			#endif
		}
	}
	return(available);
}

OSErr TLMCreateThreadPool(ThreadStyle style, short count, Size stack)
{
	if (ThreadManagerIsAvailable())
		return(CreateThreadPool(style, count, stack));
		
	return(noErr);
}

OSErr TLMGetFreeThreadCount(ThreadStyle style, short *count)
{
	if (ThreadManagerIsAvailable())
		return(GetFreeThreadCount(style, count));
		
	*count = 0;
	return(noErr);
}

OSErr TLMGetSpecificFreeThreadCount(ThreadStyle style, Size stack, short *count)
{
	if (ThreadManagerIsAvailable())
		return(GetSpecificFreeThreadCount(style, stack, count));
		
	*count = 0;
	return(noErr);
}

OSErr TLMGetDefaultThreadStackSize(ThreadStyle style, Size *size)
{
	if (ThreadManagerIsAvailable())
		return(GetDefaultThreadStackSize(style, size));
		
	*size = ThreadStackDefault();
	return(ThreadError());
}

OSErr TLMThreadCurrentStackSpace(ThreadID id, unsigned long *free)
{
	ThreadType sn;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(ThreadCurrentStackSpace(id, free));
		
	err = ThreadIDToSN(id, &sn);
	if (! err) {
		*free = ThreadStackSpace(sn);
		err = ThreadError();
	}
	return(err);
}

OSErr TLMNewThread(ThreadStyle style,
	ThreadEntryProcPtr entry,
	void *param,
	Size stack,
	ThreadOptions options,
	void **result,
	ThreadID *id)
{
	ThreadManagerType *thread;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(NewThread(style, entry, param, stack, options, result, id));
		
	err = noErr;
	thread = NULL;
	if (style == kPreemptiveThread)
		err = threadProtocolErr;
	if (! err) {
	
		/* create the main thread if it doesn't already exist */
		err = CreateMainThread();
		if (! err) {
		
			/* allocate thread structure */
			thread = (ThreadManagerType *) NewPtrClear(sizeof(ThreadManagerType));
			if (! thread)
				err = (MemError() ? MemError() : memFullErr);
			if (! err) {
			
				/* allocate thread */
				thread->entry = entry;
				thread->entryParam = param;
				thread->entryResult = result;
				thread->sn = ThreadBegin(tlm_entry, NULL, NULL, thread, stack);
				err = ThreadError();
				if (! err) {
				
					/* initialize thread */
					ThreadProcBeginSet(thread->sn, tlm_begin);
					err = ThreadError();
					if (! err) {
						ThreadProcEndSet(thread->sn, tlm_end);
						err = ThreadError();
						if (! err) {
							err = ThreadSNToID(thread->sn, &thread->id);
							if (! err) {
								if ((options & kNewSuspend) != 0) {
									ThreadEnabledSet(thread->sn, false);
									err = ThreadError();
								}
							}
						}
					}
					if (err) {
						ThreadEnd(thread->sn);
						thread->sn = THREAD_NONE;
					}
				}
				if (err) {
					DisposePtr((Ptr) thread);
					thread = NULL;
				}
			}
		}
	}
	*id = (thread ? thread->id : kNoThreadID);
	return(err);
}

OSErr TLMDisposeThread(ThreadID id, void *result, Boolean recycle)
{
	ThreadManagerType *thread;
	ThreadType sn;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(DisposeThread(id, result, recycle));
		
	err = noErr;
	if (id != kNoThreadID) {
		err = ThreadIDToSN(id, &sn);
		if (! err) {
			thread = ThreadData(sn);
			err = ThreadError();
			if (! err) {
				if (thread->entryResult && result)
					*thread->entryResult = result;
				ThreadEnd(sn);
				err = ThreadError();
				if (! err)
					DisposePtr((Ptr) thread);
			}
		}
	}
	return(err);
}

OSErr TLMGetCurrentThread(ThreadID *id)
{
	ThreadType sn;
	OSErr err;

	if (ThreadManagerIsAvailable())
		return(GetCurrentThread(id));

	sn = ThreadActive();
	err = ThreadError();
	if (! err)
		err = ThreadSNToID(sn, id);
	return(err);
}

OSErr TLMYieldToAnyThread(void)
{
	if (ThreadManagerIsAvailable())
		return(YieldToAnyThread());
		
	if (ThreadMain() != THREAD_NONE)
		ThreadYield(0);
	return(ThreadError());
}

OSErr TLMThreadBeginCritical(void)
{
	ThreadManagerType *thread;
	ThreadType sn;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(ThreadBeginCritical());

	sn = ThreadActive();
	err = ThreadError();
	if (! err) {
		thread = ThreadData(sn);
		err = ThreadError();
		if (! err) {
			if (thread->critical == LONG_MAX)
				err = threadProtocolErr;
			if (! err)
				thread->critical++;
		}
	}
	return(err);
}

OSErr TLMThreadEndCritical(void)
{
	ThreadManagerType *thread;
	ThreadType sn;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(ThreadEndCritical());

	sn = ThreadActive();
	err = ThreadError();
	if (! err) {
		thread = ThreadData(sn);
		err = ThreadError();
		if (! err) {
			if (thread->critical == 0)
				err = threadProtocolErr;
			if (! err)
				thread->critical--;
		}
	}
	return(err);
}

OSErr TLMYieldToThread(ThreadID id)
{
	ThreadType sn;
	OSErr err;

	if (ThreadManagerIsAvailable())
		return(YieldToThread(id));
		
	err = ThreadIDToSN(id, &sn);
	if (! err) {
		ThreadActivate(sn);
		err = ThreadError();
	}
	return(err);
}

OSErr TLMGetThreadState(ThreadID id, ThreadState *state)
{
	ThreadType sn;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(GetThreadState(id, state));

	err = ThreadIDToSN(id, &sn);
	if (! err) {
		if (sn == ThreadActive())
			*state = kRunningThreadState;
		else if (! ThreadEnabled(sn))
			*state = kStoppedThreadState;
		else
			*state = kReadyThreadState;
	}
	return(ThreadError());
}

OSErr TLMSetThreadState(ThreadID id, ThreadState state, ThreadID suggested)
{
	ThreadType active;
	ThreadType sn;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(SetThreadState(id, state, suggested));
	
	active = ThreadActive();
	err = ThreadError();
	if (! err) {
		err = ThreadIDToSN(id, &sn);
		if (! err) {
			switch (state) {
			case kRunningThreadState:
				if (sn != active) {
					ThreadActivate(sn);
					err = ThreadError();
				}
				break;
			case kStoppedThreadState:
				if (id == kApplicationThreadID)
					err = threadProtocolErr;
				if (! err) {
					ThreadEnabledSet(sn, false);
					err = ThreadError();
					if (! err) {
						if (sn == active)
							ThreadActivate(suggested);
					}
				}
				break;
			case kReadyThreadState:
				if (id == kApplicationThreadID)
					err = threadProtocolErr;
				if (! err) {
					ThreadEnabledSet(sn, true);
					err = ThreadError();
					if (! err) {
						if (sn == active)
							ThreadActivate(suggested);
					}
				}
				break;
			default:
				err = threadProtocolErr;
				break;
			}
		}
	}
	return(err);
}

OSErr TLMSetThreadStateEndCritical(ThreadID id, ThreadState state, ThreadID suggested)
{
	OSErr err;

	if (ThreadManagerIsAvailable())
		return(SetThreadStateEndCritical(id, state, suggested));
		
	err = TLMThreadEndCritical();
	if (! err)
		err = TLMSetThreadState(id, state, suggested);
	return(err);
}

OSErr TLMGetThreadCurrentTaskRef(ThreadTaskRef *task)
{
	if (ThreadManagerIsAvailable())
		return(GetThreadCurrentTaskRef(task));
		
	*task = (ThreadTaskRef) SetCurrentA5();
	return(noErr);
}

OSErr TLMGetThreadStateGivenTaskRef(ThreadTaskRef task, ThreadID id, ThreadState *state)
{
	long oldA5;
	OSErr err;

	if (ThreadManagerIsAvailable())
		return(GetThreadStateGivenTaskRef(task, id, state));

	oldA5 = SetA5((long) task);
	err = TLMGetThreadState(id, state);
	(void) SetA5(oldA5);
	return(err);
}

OSErr TLMSetThreadReadyGivenTaskRef(ThreadTaskRef task, ThreadID id)
{
	ThreadState state;
	long oldA5;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(SetThreadReadyGivenTaskRef(task, id));

	oldA5 = SetA5((long) task);
	err = TLMGetThreadStateGivenTaskRef(task, id, &state);
	if (! err) {
		if (state != kStoppedThreadState)
			err = threadProtocolErr;
		if (! err)
			TLMSetThreadState(id, state, kNoThreadID);
	}
	(void) SetA5(oldA5);
	return(err);
}

OSErr TLMSetThreadScheduler(ThreadSchedulerProcPtr scheduler)
{
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(SetThreadScheduler(scheduler));
		
	ThreadProcScheduleSet(scheduler ? tlm_scheduler : NULL);
	err = ThreadError();
	if (! err)
		gThread.threadScheduler = scheduler;
	return(err);
}

OSErr TLMSetThreadSwitcher(ThreadID id, ThreadSwitchProcPtr switcher,
		void *param, Boolean inout)
{
	ThreadManagerType *thread;
	ThreadType sn;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(SetThreadSwitcher(id, switcher, param, inout));
		
	err = ThreadIDToSN(id, &sn);
	if (! err) {
		thread = ThreadData(sn);
		err = ThreadError();
		if (! err) {
			if (inout) {
				ThreadProcResumeSet(sn, switcher ? tlm_resume : NULL);
				err = ThreadError();
				if (! err) {
					thread->switcherIn = switcher;
					thread->switcherInParam = param;
				}
			}
			else {
				ThreadProcSuspendSet(sn, switcher ? tlm_suspend : NULL);
				err = ThreadError();
				if (! err) {
					thread->switcherOut = switcher;
					thread->switcherOutParam = param;
				}
			}
		}
	}
	return(err);
}

OSErr TLMSetThreadTerminator(ThreadID id, ThreadSwitchProcPtr terminator,
		void *param)
{
	ThreadManagerType *thread;
	ThreadType sn;
	OSErr err;
	
	if (ThreadManagerIsAvailable())
		return(SetThreadTerminator(id, terminator, param));
	
	if (id == kApplicationThreadID)
		err = threadProtocolErr;
	else {
		err = ThreadIDToSN(id, &sn);
		if (! err) {
			thread = ThreadData(sn);
			err = ThreadError();
			if (! err) {
				ThreadProcEndSet(sn, terminator ? tlm_end : NULL);
				err = ThreadError();
				if (! err) {
					thread->terminator = terminator;
					thread->terminatorParam = param;
				}
			}
		}
	}
	return(err);
}
